"""Test `radiative_forcing.py` module."""

import numpy as np
import pytest

from pycontrails.models.cocip import CocipParams, radiative_forcing
from pycontrails.models.cocip.radiative_forcing import RFConstants, rf_const
from pycontrails.physics import geo


def test_rf_const() -> None:
    """Test the RFConstant types."""

    assert isinstance(rf_const, RFConstants)

    # the indexes of the habit_weights should be the (radius_threshold x habits)
    # includes radius < rf_const.radius_threshold_um[0] as the 0th row index
    cp = CocipParams()
    assert cp.habit_distributions.shape == (cp.radius_threshold_um.size + 1, cp.habits.size)

    # make sure habit specific attributes are the same length as the habits
    for attr, val in RFConstants.__dict__.items():
        if not attr.startswith("_"):
            assert isinstance(val, np.ndarray)
            assert val.size == cp.habits.size


def test_habit_weights() -> None:
    """Test habit weight classification."""

    r_vol_um = np.linspace(1, 500, 100)
    cp = CocipParams()
    habit_distributions = cp.habit_distributions
    radius_threshold_um = cp.radius_threshold_um
    habits = cp.habits

    # calculate habit weight distribution regime based on particle size
    regimes = radiative_forcing.habit_weight_regime_idx(r_vol_um, radius_threshold_um)

    # must correspond to index in `habits`
    assert (regimes < 7).all()

    # should all be increasing as the particle size monotonically increases
    assert (np.diff(regimes) >= 0).all()

    # habit_weights should provide a weight value for each waypoint
    weights = radiative_forcing.habit_weights(r_vol_um, habit_distributions, radius_threshold_um)
    assert weights.shape == (r_vol_um.size, habits.size)


def test_effective_radius() -> None:
    """Test effective radius calculations."""

    r_vol_um = np.linspace(1, 350, 100)
    cp = CocipParams()
    habit_distributions = cp.habit_distributions
    radius_threshold_um = cp.radius_threshold_um

    weights = radiative_forcing.habit_weights(r_vol_um, habit_distributions, radius_threshold_um)
    habit_weight_idxs = np.where(weights > 0)
    r_vol_um_h = r_vol_um[habit_weight_idxs[0]]
    habit_types_idx = habit_weight_idxs[1]

    r_eff_um = radiative_forcing.effective_radius_by_habit(r_vol_um_h, habit_types_idx)

    # all effective radii should be > 0 and with 0.5 and 1.5 of particle size
    # TODO: many values saturate at 45um?
    assert np.all(r_eff_um > 0)
    assert np.all(r_eff_um < 1.5 * r_vol_um_h)
    assert np.all((r_eff_um > 0.4 * r_vol_um_h) | (r_eff_um == 45))


def test_override_effective_radius() -> None:
    """Allow override of r_eff_um in methods."""

    r_vol_um = np.linspace(1, 350, 100)
    r_eff_um = r_vol_um * 0.9

    cp = CocipParams()

    habit_distributions = cp.habit_distributions.astype("float64")
    radius_threshold_um = cp.radius_threshold_um.astype("float64")

    # either floats or the same shape as r_vol_um
    latitude = np.full(r_vol_um.shape, 43)
    longitude = np.full(r_vol_um.shape, -73)
    time = np.full(r_vol_um.shape, np.datetime64("2022-04-01 12:00:00"))

    air_temperature = np.full(r_vol_um.shape, 235)
    tau_contrail = np.full(r_vol_um.shape, 1)  # 0 - 3
    tau_cirrus = np.full(r_vol_um.shape, 0)  # 0 - 10

    top_net_solar_radiation = np.full(r_vol_um.shape, 100)
    top_net_thermal_radiation = np.full(r_vol_um.shape, -200)

    # calculate instantaneous theoretical solar direct radiation based on geo position and time
    theoretical_sdr = geo.solar_direct_radiation(longitude, latitude, time)
    theta_rad = geo.orbital_position(time)
    sd0 = geo.solar_constant(theta_rad)

    # calculate the average reflected shortwave radiation from theoretical sdr and tnsr
    rsr = np.maximum(theoretical_sdr - top_net_solar_radiation, 0)

    # calculate the average outgoing longwave radiation from tnsr
    olr = -1 * top_net_thermal_radiation

    # radiation calculations
    habit_weights = radiative_forcing.habit_weights(
        r_vol_um, habit_distributions, radius_threshold_um
    )

    rf_lw = radiative_forcing.longwave_radiative_forcing(
        r_vol_um, olr, air_temperature, tau_contrail, tau_cirrus, habit_weights
    )
    rf_lw_2 = radiative_forcing.longwave_radiative_forcing(
        r_vol_um, olr, air_temperature, tau_contrail, tau_cirrus, habit_weights, r_eff_um=r_eff_um
    )

    rf_sw = radiative_forcing.shortwave_radiative_forcing(
        r_vol_um, theoretical_sdr, rsr, sd0, tau_contrail, tau_cirrus, habit_weights
    )
    rf_sw_2 = radiative_forcing.shortwave_radiative_forcing(
        r_vol_um,
        theoretical_sdr,
        rsr,
        sd0,
        tau_contrail,
        tau_cirrus,
        habit_weights,
        r_eff_um=r_eff_um,
    )

    # values shouldn't be the same, but should only be off by < 1
    # FIXME: I think some value differences are just floating point errors
    assert (rf_lw != rf_lw_2).all()
    assert (np.abs(rf_lw - rf_lw_2) < 1).all()

    # values shouldn't be the same, but should only be off by < 1
    assert (rf_sw != rf_sw_2).all()
    assert (np.abs(rf_sw - rf_sw_2) < 1).all()


def test_override_habit_distributions() -> None:
    """Test override habit distributions."""
    r_vol_um = np.linspace(1, 500, 100)

    cp = CocipParams()

    habit_distributions = cp.habit_distributions
    radius_threshold_um = cp.radius_threshold_um
    habits = cp.habits

    habit_distributions[1, 1] = 0.0
    habit_distributions[1, 2] = 0.3

    # radiation calculations
    habit_weights = radiative_forcing.habit_weights(
        r_vol_um, habit_distributions, radius_threshold_um
    )
    assert habit_weights.shape == (r_vol_um.size, habits.size)
    assert habit_weights[1, 2] == pytest.approx(0.3)
    assert habit_weights[1, 1] != 0.3

    with pytest.raises(ValueError, match="distributions must sum to 1"):
        habit_distributions[1, 2] = 1
        radiative_forcing.habit_weights(r_vol_um, habit_distributions, radius_threshold_um)

    with pytest.raises(ValueError, match="number of rows in"):
        habit_distributions[1, 2] = 0.3
        radiative_forcing.habit_weights(r_vol_um, habit_distributions, np.array([5, 10]))



"""
Output benchmark test results.

```
$ python output.py
```
"""

import logging
import pathlib

import benchmark

# logging set up in benchmark.py
LOG = logging.getLogger("pycontrails")


LOG.info("Running Cocip over flights")

# run model on benchmark inputs
cocip = benchmark.run_cocip()

LOG.info("Writing outputs")

# -------------
# Flight output
# -------------

fleet = benchmark.parse_flight_results(cocip.source)

# output one file per flight id
pathlib.Path(benchmark.OUTPUT_PATH / "flight").mkdir(parents=True, exist_ok=True)
for fid, df in fleet.groupby("flight_id"):
    # prefer parquet
    df.reset_index(drop=True).to_parquet(
        benchmark.OUTPUT_PATH / "flight" / f"{fid}.pq", index=False
    )

    # # Output csv
    # # fill na with -9999
    # df.reset_index(drop=True) \
    #   .fillna(value=-9999) \
    #   .to_csv(benchmark.OUTPUT_PATH / "flight" / f"{fid}.csv", index=False)


# ---------------
# Contrail output
# ---------------

contrail = benchmark.parse_contrail_results(cocip.contrail)

# output one file per flight id
pathlib.Path(benchmark.OUTPUT_PATH / "contrail").mkdir(parents=True, exist_ok=True)
for fid, df in contrail.groupby("flight_id"):
    # prefer parquet
    df.reset_index(drop=True).to_parquet(
        benchmark.OUTPUT_PATH / "contrail" / f"{fid}.pq", index=False
    )

    # # Output csv
    # # fill na with -9999
    # df.reset_index(drop=True) \
    #   .fillna(value=-9999) \
    #   .to_csv(benchmark.OUTPUT_PATH / "contrail" / f"{fid}.csv", index=False)



"""Test pycontrails.core.coordinates."""

from __future__ import annotations

import numpy as np
import pandas as pd
import pytest

from pycontrails.core import coordinates


def test_slice_domain() -> None:
    """Test coordinates.slice_domain()."""
    domain = np.arange(-180, 180, 1.0)

    # normal request within bounds
    request = np.arange(-20, 20, 1)
    sl = coordinates.slice_domain(domain, request)
    assert isinstance(sl, slice)
    assert sl.start == 160
    assert sl.stop == 200
    assert sl.step is None

    # normal request within bounds with buffer
    request = np.arange(-20, 20, 1)
    sl = coordinates.slice_domain(domain, request, buffer=(5, 10))
    assert isinstance(sl, slice)
    assert sl.start == 155
    assert sl.stop == 210
    assert sl.step is None

    # on the min edge of the bounds
    request = np.arange(-200, 20, 1)
    sl = coordinates.slice_domain(domain, request)
    assert sl.start == 0
    assert sl.stop == 200
    assert sl.step is None

    # complete outside min edge of the bounds
    request = np.arange(-220, -190, 1)
    sl = coordinates.slice_domain(domain, request)
    assert sl.start == 0
    assert sl.stop == 2
    assert sl.step is None

    # on the max edge of the bounds
    request = np.arange(-20, 200, 1)
    sl = coordinates.slice_domain(domain, request)
    assert sl.start == 160
    assert sl.stop == len(domain)
    assert sl.step is None

    # completely outside min edge of the bounds
    request = np.arange(190, 220, 1)
    sl = coordinates.slice_domain(domain, request)
    assert sl.start == len(domain) - 2
    assert sl.stop == len(domain)
    assert sl.step is None

    # handle nans
    request = np.arange(-20.0, 20.0, 1)
    request[3] = np.nan
    sl = coordinates.slice_domain(domain, request)
    assert sl.start == 160
    assert sl.stop == 200
    assert sl.step is None

    # fake domain that is short should just be returned
    domain = np.array([0, 1])
    request = np.arange(100, 200, 1)
    sl = coordinates.slice_domain(domain, request)
    assert sl.start is None
    assert sl.stop is None
    assert sl.step is None

    # unsorted domain should throw an error
    request = np.arange(-20, 20, 1)
    with pytest.raises(ValueError, match="Domain must be sorted"):
        sl = coordinates.slice_domain(np.array([3, 1, 2, 4, 5, 6]), request)


def test_buffer_domain_time_like():
    """Test coordinates.slice_domain on time-like domain."""
    domain = pd.date_range("2022-02-01", "2022-02-02", 321)
    request = np.datetime64("2022-02-01T06"), np.datetime64("2022-02-01T12")
    buffer = np.timedelta64(1, "h"), np.timedelta64(1, "h")
    sl = coordinates.slice_domain(domain, request, buffer)
    assert sl.start == 66
    assert sl.stop == 175

    with pytest.warns(UserWarning):
        buffer = buffer[0], -buffer[1]
        coordinates.slice_domain(domain, request, buffer)

    with pytest.warns(UserWarning):
        buffer = -buffer[0], buffer[1]
        coordinates.slice_domain(domain, request, buffer)


def test_buffer_domain_warning():
    """Confirm that slice_domain gives a warning for a negative buffer.

    (This has burned me in the past!)
    """
    domain = np.arange(-180, 180, 1.0)
    request = -100.5, 100.5
    buffer = 0, 0
    sl = coordinates.slice_domain(domain, request, buffer)
    np.testing.assert_array_equal(domain[sl], domain[(domain <= 101) & (domain >= -101)])

    with pytest.warns(UserWarning):
        buffer = -1, 0
        coordinates.slice_domain(domain, request, buffer)

    with pytest.warns(UserWarning):
        buffer = 3, -1
        coordinates.slice_domain(domain, request, buffer)


def test_intersect_domain() -> None:
    """Test coordinates.intersect_domain()."""
    request = np.arange(-20, 20, 1.0)

    # all request points are within domain
    domain = np.arange(-30, 30, 1.0)
    mask = coordinates.intersect_domain(domain, request)
    assert isinstance(mask, np.ndarray)
    assert mask.all()

    # some domain points are within request
    domain = np.arange(-30, 10, 1.0)
    mask = coordinates.intersect_domain(domain, request)
    assert mask.any()
    assert not mask[31:40].any()

    domain = np.arange(-10, 10, 1.0)
    mask = coordinates.intersect_domain(domain, request)
    assert mask.any()
    assert not mask[0:10].any() and not mask[31:40].any()

    domain = np.arange(10, 30, 1.0)
    mask = coordinates.intersect_domain(domain, request)
    assert mask.any()
    assert not mask[0:30].any()

    # no domain points are within request
    domain = np.arange(-40, -25, 1.0)
    mask = coordinates.intersect_domain(domain, request)
    assert not mask.any()

    # supports unsorted requests
    request_unsorted = np.array([3, 1, 2, 4, 5, 10, -1, -5, 3])
    domain = np.arange(-30, 30, 1.0)
    mask = coordinates.intersect_domain(domain, request_unsorted)
    assert mask.all()

    domain = np.arange(-30, 0, 1.0)
    mask = coordinates.intersect_domain(domain, request_unsorted)
    assert mask[6:8].all() and not mask[0:6].any() and not mask[8].any()



"""Common utilities for ECMWF Data Access."""

from __future__ import annotations

import logging
from typing import Any

LOG = logging.getLogger(__name__)

import numpy as np
import xarray as xr

from pycontrails.core import datalib, met


def rad_accumulated_to_average(mds: met.MetDataset, key: str, dt_accumulation: int) -> None:
    """Convert accumulated radiation value to instantaneous average.

    Parameters
    ----------
    mds : MetDataset
        MetDataset containing the accumulated value at ``key``
    key : str
        Data variable key
    dt_accumulation : int
        Accumulation time in seconds, [:math:`s`]
    """
    if key in mds.data and not mds.data[key].attrs.get("_pycontrails_modified", False):
        if not np.all(np.diff(mds.data["time"]) == np.timedelta64(dt_accumulation, "s")):
            raise ValueError(
                f"Dataset expected to have time interval of {dt_accumulation} seconds when"
                " converting accumulated parameters"
            )

        mds.data[key] = mds.data[key] / dt_accumulation
        mds.data[key].attrs["units"] = "W m**-2"
        mds.data[key].attrs[
            "_pycontrails_modified"
        ] = "Accumulated value converted to average instantaneous value"


# TODO: Remove this in favor of functional implementation
class ECMWFAPI(datalib.MetDataSource):
    """Abstract class for all ECMWF data accessed remotely through CDS / MARS."""

    @property
    def variable_ecmwfids(self) -> list[int]:
        """Return a list of variable ecmwf_ids.

        Returns
        -------
        list[int]
            List of int ECMWF param ids.
        """
        return [v.ecmwf_id for v in self.variables if v.ecmwf_id is not None]

    # TODO: this could be functional, but there many properties utilized
    def _process_dataset(self, ds: xr.Dataset, **kwargs: Any) -> met.MetDataset:
        """Process the :class:`xr.Dataset` opened from cache or local files.

        Parameters
        ----------
        ds : xr.Dataset
            Dataset loaded from netcdf cache files or input paths.
        **kwargs : Any
            Keyword arguments passed through directly into :class:`MetDataset` constructor.

        Returns
        -------
        MetDataset
        """

        # downselect variables
        try:
            ds = ds[self.variable_shortnames]
        except KeyError as e:
            raise KeyError(f"Input dataset is missing variables {e}")

        # downselect times
        if not self.timesteps:
            self.timesteps = ds["time"].values.astype("datetime64[ns]").tolist()
        else:
            try:
                ds = ds.sel(time=self.timesteps)
            except KeyError:
                # this snippet shows the missing times for convenience
                np_timesteps = [np.datetime64(t, "ns") for t in self.timesteps]
                missing_times = sorted(set(np_timesteps) - set(ds["time"].values))
                raise KeyError(
                    f"Input dataset is missing time coordinates {[str(t) for t in missing_times]}"
                )

        # downselect pressure level
        # if "level" is not in dims and
        # length of the requested pressure levels is 1
        # expand the dims with this level
        if "level" not in ds.dims and len(self.pressure_levels) == 1:
            ds = ds.expand_dims({"level": self.pressure_levels})

        try:
            ds = ds.sel(level=self.pressure_levels)
        except KeyError:
            # this snippet shows the missing levels for convenience
            missing_levels = sorted(set(self.pressure_levels) - set(ds["level"].values))
            raise KeyError(f"Input dataset is missing level coordinates {missing_levels}")

        # harmonize variable names
        ds = met.standardize_variables(ds, self.variables)

        # modify values

        # rescale relative humidity from % -> dimensionless if its in dataset
        if "relative_humidity" in ds and not ds["relative_humidity"].attrs.get(
            "_pycontrails_modified", False
        ):
            ds["relative_humidity"] = ds["relative_humidity"] / 100
            ds["relative_humidity"].attrs["long_name"] = "Relative humidity"
            ds["relative_humidity"].attrs["standard_name"] = "relative_humidity"
            ds["relative_humidity"].attrs["units"] = "[0 - 1]"
            ds["relative_humidity"].attrs[
                "_pycontrails_modified"
            ] = "Relative humidity rescaled to [0 - 1] instead of %"

        ds.attrs["met_source"] = type(self).__name__

        kwargs.setdefault("cachestore", self.cachestore)
        return met.MetDataset(ds, **kwargs)



"""Thermodynamic relationships."""

from __future__ import annotations

import numpy as np

from pycontrails.physics import constants
from pycontrails.utils.types import ArrayScalarLike, support_arraylike

# -------------------
# Material Properties
# -------------------


def rho_d(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate air density for (T, p) assuming dry air.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Air density of dry air, [:math:`kg \ m^{-3}`]
    """
    return p / (constants.R_d * T)


def rho_v(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate the air density for (T, p) assuming all water vapor.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Air density of water vapor, [:math:`kg \ m^{-3}`]
    """
    return p / (constants.R_v * T)


def c_pm(q: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate isobaric heat capacity of moist air.

    Parameters
    ----------
    q : ArrayScalarLike
        Specific humidity, [:math:`kg \ kg^{-1}`]

    Returns
    -------
    ArrayScalarLike
        Isobaric heat capacity of moist air, [:math:`J \ kg^{-1} \ K^{-1}`]

    Notes
    -----
    Some models (including CoCiP) use a constant value here (1004 :math:`J \ kg^{-1} \ K^{-1}`)

    """
    return constants.c_pd * (1 + (constants.c_pv / constants.c_pd + 1) * q)


def p_vapor(q: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate the vapor pressure.

    Parameters
    ----------
    q : ArrayScalarTypeVar
        Specific humidity, [:math:`kg \ kg^{-1}`]
    p : ArrayScalarTypeVar
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarTypeVar
        Vapor pressure, [:math:`Pa`]
    """
    return q * p * (constants.R_v / constants.R_d)


# -------------------
# Saturation Pressure
# -------------------


def e_sat_ice(T: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate saturation pressure of water vapor over ice.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]

    Returns
    -------
    ArrayScalarLike
        Saturation pressure of water vapor over ice, [:math:`Pa`]

    References
    ----------
    - :cite:`sonntag1994`

    """
    # Goff Gratch equation (Smithsonian Tables, 1984)
    # return np.log10(-9.09718 * (273.16/T - 1) - 3.56654 * np.log10(273.16/T) + \
    #                  0.87679 * (1 - T/273.16) + np.log10(6.1071))

    # Magnus Teten (Murray, 1967)
    # return 6.1078 * np.exp(21.8745 * (T - 273.16) / (T - 7.66))

    # Zhang 2017 - incorrect implementation of Magnus Teten
    # return 6.1808 * np.exp(21.875 * (T - 276.16) / (T - 7.66))

    # Guide to Meteorological Instruments and Methods of Observation (CIMO Guide) (WMO, 2008)
    # return 6.112 * np.exp(22.46 * (T - 273.16) / (272.62 + T - 273.16))

    # Sonntag (1994) is used in CoCiP

    # FIXME: Presently, mypy is not aware that numpy ufuncs will return `xr.DataArray``
    # when xr.DataArray is passed in. This will get fixed at some point in the future
    # as `numpy` their typing patterns, after which the "type: ignore" comment can
    # get ripped out.
    # We could explicitly check for `xr.DataArray` then use `xr.apply_ufunc`, but
    # this only renders our code more boilerplate and less performant.
    # This comment is pasted several places in `pycontrails` -- they should all be
    # addressed at the same time.
    return 100 * np.exp(  # type: ignore[return-value]
        (-6024.5282 / T)
        + 24.7219
        + (0.010613868 * T)
        - (1.3198825e-5 * (T**2))
        - 0.49382577 * np.log(T)
    )


def e_sat_liquid(T: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate saturation pressure of water vapor over liquid water.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]

    Returns
    -------
    ArrayScalarLike
        Saturation pressure of water vapor over liquid water, [:math:`Pa`]

    References
    ----------
    - :cite:`sonntag1994`
    """
    # Buck (Buck Research Manual 1996)
    # 6.1121 * np.exp((18.678 * (T - 273.15) / 234.5) * (T - 273.15) / (257.14 + (T - 273.15)))

    # Magnus Tetens (Murray, 1967)
    # 6.1078 * np.exp(17.269388 * (T - 273.16) / (T – 35.86))

    # Guide to Meteorological Instruments and Methods of Observation (CIMO Guide) (WMO, 2008)
    # 6.112 * np.exp(17.62 * (T - 273.15) / (243.12 + T - 273.15))

    # Sonntag (1994) is used in CoCiP

    # FIXME: Presently, mypy is not aware that numpy ufuncs will return `xr.DataArray``
    # when xr.DataArray is passed in. This will get fixed at some point in the future
    # as `numpy` their typing patterns, after which the "type: ignore" comment can
    # get ripped out.
    # We could explicitly check for `xr.DataArray` then use `xr.apply_ufunc`, but
    # this only renders our code more boilerplate and less performant.
    # This comment is pasted several places in `pycontrails` -- they should all be
    # addressed at the same time.
    return 100 * np.exp(  # type: ignore[return-value]
        -6096.9385 / T
        + 16.635794
        - 0.02711193 * T
        + 1.673952 * 1e-5 * T**2
        + 2.433502 * np.log(T)
    )


@support_arraylike
def _e_sat_piecewise(T: np.ndarray) -> np.ndarray:
    """Calculate `e_sat_liquid` when T is above freezing otherwise `e_sat_ice`.

    Parameters
    ----------
    T : np.ndarray
        Temperature, [:math:`K`]

    Returns
    -------
    np.ndarray
        Piecewise array of e_sat_liquid and e_sat_ice values.
    """
    condlist = [T >= -constants.absolute_zero, T < constants.absolute_zero]
    funclist = [e_sat_liquid, e_sat_ice, np.nan]  # nan passed through
    return np.piecewise(T, condlist, funclist)


# ----------------------------
# Saturation Specific Humidity
# ----------------------------


def q_sat(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate saturation specific humidity over liquid or ice.

    When T is above 0 C, liquid saturation is computed. Otherwise, ice saturation
    is computed.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Saturation specific humidity, [:math:`kg \ kg^{-1}`]

    Notes
    -----
    Smith et al. (1999)
    """
    e_sat = _e_sat_piecewise(T)
    return constants.epsilon * e_sat / p


def q_sat_ice(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate saturation specific humidity over ice.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Saturation specific humidity, [:math:`kg \ kg^{-1}`]

    Notes
    -----
    Smith et al. (1999)
    """
    return constants.epsilon * e_sat_ice(T) / p


def q_sat_liquid(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate saturation specific humidity over liquid.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Saturation specific humidity, [:math:`kg \ kg^{-1}`]

    Notes
    -----
    Smith et al. (1999)
    """
    return constants.epsilon * e_sat_liquid(T) / p


# -----------------
# Relative Humidity
# -----------------


def rh(q: ArrayScalarLike, T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate the relative humidity with respect to to liquid water.

    Parameters
    ----------
    q : ArrayScalarLike
        Specific humidity, [:math:`kg \ kg^{-1}`]
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Relative Humidity, :math:`[0 - 1]`
    """
    return (q * p * (constants.R_v / constants.R_d)) / e_sat_liquid(T)


def rhi(q: ArrayScalarLike, T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate the relative humidity with respect to ice (RHi).

    Parameters
    ----------
    q : ArrayScalarLike
        Specific humidity, [:math:`kg \ kg^{-1}`]
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Relative Humidity over ice, :math:`[0 - 1]`
    """
    return (q * p * (constants.R_v / constants.R_d)) / e_sat_ice(T)


# --------------
# Met Properties
# --------------


def p_dz(T: ArrayScalarLike, p: ArrayScalarLike, dz: float) -> ArrayScalarLike:
    r"""Calculate the pressure altitude ``dz`` meters below input pressure.

    Returns surface pressure if the calculated pressure altitude is greater
    than :const:`constants.p_surface`.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature, [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]
    dz : float
        Difference in altitude between measurements, [:math:`m`]

    Returns
    -------
    ArrayScalarLike
        Pressure at altitude, [:math:`Pa`]

    Notes
    -----
    This is used to calculate the temperature gradient and wind shear.
    """
    dp = rho_d(T, p) * constants.g * dz

    # FIXME: Presently, mypy is not aware that numpy ufuncs will return `xr.DataArray``
    # when xr.DataArray is passed in. This will get fixed at some point in the future
    # as `numpy` their typing patterns, after which the "type: ignore" comment can
    # get ripped out.
    # We could explicitly check for `xr.DataArray` then use `xr.apply_ufunc`, but
    # this only renders our code more boilerplate and less performant.
    # This comment is pasted several places in `pycontrails` -- they should all be
    # addressed at the same time.
    return np.minimum(p + dp, constants.p_surface)  # type: ignore[return-value]


def T_potential_gradient(
    T_top: ArrayScalarLike,
    p_top: ArrayScalarLike,
    T_btm: ArrayScalarLike,
    p_btm: ArrayScalarLike,
    dz: float,
) -> ArrayScalarLike:
    r"""Calculate the potential temperature gradient between two altitudes.

    Parameters
    ----------
    T_top : ArrayScalarLike
        Temperature at original altitude, [:math:`K`]
    p_top : ArrayScalarLike
        Pressure at original altitude, [:math:`Pa`]
    T_btm : ArrayScalarLike
        Temperature at lower altitude, [:math:`K`]
    p_btm : ArrayScalarLike
        Pressure at lower altitude, [:math:`Pa`]
    dz : float
        Difference in altitude between measurements, [:math:`m`]

    Returns
    -------
    ArrayScalarLike
        Potential Temperature gradient, [:math:`K \ m^{-1}`]
    """
    T_potential_top = T_potential(T_top, p_top)
    T_potential_btm = T_potential(T_btm, p_btm)
    return (T_potential_top - T_potential_btm) / dz


def T_potential(T: ArrayScalarLike, p: ArrayScalarLike) -> ArrayScalarLike:
    r"""Calculate potential temperature.

    The potential temperature is the temperature that
    an air parcel would attain if adiabatically
    brought to a standard reference pressure, :const:`constants.p_surface`.

    Parameters
    ----------
    T : ArrayScalarLike
        Temperature , [:math:`K`]
    p : ArrayScalarLike
        Pressure, [:math:`Pa`]

    Returns
    -------
    ArrayScalarLike
        Potential Temperature, [:math:`K`]

    References
    ----------
    - https://en.wikipedia.org/wiki/Potential_temperature
    """
    return T * (constants.p_surface / p) ** (constants.R_d / constants.c_pd)


def brunt_vaisala_frequency(p: np.ndarray, T: np.ndarray, T_grad: np.ndarray) -> np.ndarray:
    r"""Calculate the Brunt-Vaisaila frequency.

    The Brunt-Vaisaila frequency is the frequency at which a vertically
    displaced parcel will oscillate within a statically stable environment.

    Parameters
    ----------
    p : np.ndarray
        Pressure, [:math:`Pa`]
    T : np.ndarray
        Temperature , [:math:`K`]
    T_grad : np.ndarray
        Potential Temperature gradient (see :func:`T_potential_gradient`), [:math:`K \ m^{-1}`]

    Returns
    -------
    np.ndarray
        Brunt-Vaisaila frequency, [:math:`s^{-1}`]

    References
    ----------
    - https://en.wikipedia.org/wiki/Brunt%E2%80%93V%C3%A4is%C3%A4l%C3%A4_frequency
    """
    theta = T_potential(T, p)
    T_grad.clip(min=1e-6, out=T_grad)
    return (T_grad * constants.g / theta) ** 0.5



"""Test `MetDataset` opened from a Zarr store.

These tests require a ERA5_PL_ZARR_STORE environment variable and an internet
connection. They re slow to run.
"""

import os

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pycontrails import GeoVectorDataset, MetDataArray, MetDataset

ZARR_STORE = os.getenv("ERA5_PL_ZARR_STORE")

if ZARR_STORE is None:
    pytest.skip("Zarr store not available from environment variable", allow_module_level=True)


@pytest.fixture()
def ds():
    ds_ = xr.open_zarr(ZARR_STORE)  # using default chunks="auto" for dask interop
    try:
        # This raises ValueError if chunks needs work
        # This seems to be happening only for an auxiliary coordinate.
        ds_.chunks
        return ds_
    except ValueError:
        return ds_.unify_chunks()  # cheap if only a single time-derived coordinate is fixed


def test_open_zarr_as_dataset(ds: xr.Dataset):
    """Confirm interaction with a zarr-based Dataset.

    This test downloads a single chunks from the Zarr store.
    """
    assert isinstance(ds, xr.Dataset)
    da = ds["eastward_wind"]
    assert isinstance(da, xr.DataArray)
    point = da.sel(longitude=55, latitude=33, level=350, time="2020-09-02T13")
    assert point.size == 1

    # This will download the right chunk from the cloud then send it through the
    # the zarr -> dask -> xarray pipeline
    assert point.values.item() == 13.631146430969238


def test_open_zarr_as_metdataset(ds: xr.Dataset):
    """Confirm a zarr-based Dataset can be opened as a MetDataset.

    No chunks (apart from xarray dimensions) are actually downloaded here.
    """
    mds = MetDataset(ds)
    assert isinstance(mds, MetDataset)
    assert isinstance(mds.shape, tuple)
    assert mds.is_wrapped
    assert mds.is_zarr

    mda = mds["air_temperature"]
    assert isinstance(mda, MetDataArray)

    assert mda.is_zarr
    assert mda.is_wrapped
    assert not mda.in_memory


def test_zarr_interpolation():
    """Confirm a zarr-based Dataset can interpolate a GeoVectorDataset.

    Two time chunks are downloaded here.
    """
    mds = MetDataset.from_zarr(ZARR_STORE, cache_size=1e9)
    mda = mds["eastward_wind"]

    rng = np.random.default_rng(13579)
    n = 10000
    longitude = rng.uniform(-180, 180, n)
    latitude = rng.uniform(-90, 90, n)
    level = rng.uniform(200, 500, n)
    time = pd.date_range("2020-03-14T15", "2020-03-14T16", n)

    with pytest.raises(RuntimeError, match=r"loading at least \d+ GB of data into memory"):
        mda.interpolate(longitude, latitude, level, time)

    vector = GeoVectorDataset(longitude=longitude, latitude=latitude, level=level, time=time)
    mds = vector.downselect_met(mds)
    assert len(mds.variables["time"]) == 2
    assert np.all(vector.coords_intersect_met(mds))

    # Run interpolation on downselected data
    out = vector.intersect_met(mds["air_temperature"])
    assert isinstance(out, np.ndarray)
    assert out.size == n
    assert np.all(np.isfinite(out))
    assert np.all(out < 273)
    assert np.all(out > 195)



"""
Build netcdf assets for cocip unit tests.

- Requires account with [Copernicus Data Portal](https://cds.climate.copernicus.eu/cdsapp#!/home)
and  local `~/.cdsapirc` file with credentials.
"""

import pathlib
from datetime import datetime

from pycontrails import DiskCacheStore
from pycontrails.datalib.ecmwf import ERA5
from pycontrails.models import Cocip

# use separate disk cache
disk_cache = DiskCacheStore(cache_dir="cache")

# paths
parent = pathlib.Path(__file__).parents[1].absolute()
static_dir = parent / "unit" / "static"

## Set Domain
time = (datetime(2019, 1, 1, 0), datetime(2019, 1, 1, 12))
pressure_levels = [300, 250, 225, 200]

## Get Data
variables = (*Cocip.met_variables, "fraction_of_cloud_cover")
era5pl = ERA5(time, variables, pressure_levels, cachestore=disk_cache)
era5sl = ERA5(time, Cocip.rad_variables, cachestore=disk_cache)

# download data from ERA5 (or open from cache)
met = era5pl.open_metdataset(chunks={"time": 1, "level": 2})
rad = era5sl.open_metdataset(chunks={"time": 1})

## Write out for tests

met.data = met.data[
    dict(
        latitude=((met.data["latitude"] < 60) & (met.data["latitude"] > 50)),
        longitude=((met.data["longitude"] < -20) & (met.data["longitude"] > -40)),
    )
]
rad.data = rad.data[
    dict(
        latitude=((rad.data["latitude"] < 60) & (rad.data["latitude"] > 50)),
        longitude=((rad.data["longitude"] < -20) & (rad.data["longitude"] > -40)),
    )
]


step = 5
met.data = met.data.reindex(
    latitude=met.data["latitude"].values[0::step], longitude=met.data["longitude"].values[0::step]
)
rad.data = rad.data.reindex(
    latitude=rad.data["latitude"].values[0::step], longitude=rad.data["longitude"].values[0::step]
)

# outputs
met.data.to_netcdf(static_dir / "met-era5-cocip1.nc")
rad.data.to_netcdf(static_dir / "rad-era5-cocip1.nc")


# ## Test import

# met = MetDataset(xr.open_dataset("met-era5-cocip1.nc"))
# rad = MetDataset(xr.open_dataset("rad-era5-cocip1.nc"))



"""Test functions in the `Humidity_scaling` module."""

from __future__ import annotations

import json
from typing import Any

import numpy as np
import pandas as pd
import pytest
from overrides import overrides

from pycontrails import GeoVectorDataset, VectorDataset
from pycontrails.models import humidity_scaling as hs
from pycontrails.physics import constants, thermo, units
from pycontrails.utils.json import NumpyEncoder
from pycontrails.utils.types import ArrayLike

cls_list = [
    cls
    for cls in hs.__dict__.values()
    if isinstance(cls, type)
    and issubclass(cls, hs.HumidityScaling)
    and cls is not hs.HumidityScaling
    and cls is not hs.HistogramMatchingWithEckel
]


@pytest.fixture(scope="module")
def rng():
    """Get random number generator for module."""
    return np.random.default_rng(1234)


@pytest.fixture
def vector(rng: np.random.Generator):
    """Create vector on which to run humidity scaling."""
    n = 100
    longitude = rng.uniform(-180, 180, n)
    latitude = rng.uniform(-90, 90, n)
    altitude = rng.uniform(5000, 12000, n)
    time = pd.date_range("2019-01-01", "2019-02-01", n)  # never used
    vector = GeoVectorDataset(longitude=longitude, latitude=latitude, altitude=altitude, time=time)
    T = vector["air_temperature"] = units.m_to_T_isa(altitude)
    rhi = rng.triangular(0.3, 0.7, 1.3, n)
    rhi_on_q = vector.air_pressure * (constants.R_v / constants.R_d) / thermo.e_sat_ice(T)
    vector["specific_humidity"] = rhi / rhi_on_q
    return vector


class DefaultHumidityScaling(hs.HumidityScaling):
    """Compute uncorrected RHi without any scaling.

    This is simply a wrapper around :func:`thermo.rhi` to ensure a consistent
    humidity scaling interface.
    """

    name = "no_scaling"
    long_name = "No humidity scaling"
    formula = "rhi -> rhi"

    @overrides
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: Any,
    ) -> tuple[ArrayLike, ArrayLike]:
        rhi = thermo.rhi(specific_humidity, air_temperature, air_pressure)
        return specific_humidity, rhi


@pytest.mark.parametrize("scaler_cls", cls_list)
def test_all_scalers(vector: GeoVectorDataset, scaler_cls: type):
    """Check basic usage of humidity scaler instances."""
    scaler = scaler_cls(copy_source=False)
    q1 = vector["specific_humidity"]
    assert "rhi" not in vector

    # Call the scaler and check how vector has been mutated
    scaler.eval(vector)
    assert "rhi" in vector
    q2 = vector["specific_humidity"]

    rhi1 = thermo.rhi(vector["specific_humidity"], vector["air_temperature"], vector.air_pressure)
    rhi2 = vector["rhi"]
    np.testing.assert_allclose(rhi1, rhi2, rtol=1e-15, atol=1e-15)

    # Humidity adjusted from the original
    assert np.any(q1 != q2)

    # But not by too much
    np.testing.assert_allclose(q1, q2, rtol=0.26)


def test_rhi_adj_override(vector: GeoVectorDataset):
    """Confirm vector variables override scaler params."""
    rhi0 = thermo.rhi(vector["specific_humidity"], vector["air_temperature"], vector.air_pressure)

    # instantiate scaler
    scaler = hs.ConstantHumidityScaling(rhi_adj=0.9)

    # use scaler rhi_adj
    vector1 = scaler.eval(vector)
    rhi1 = vector1["rhi"]

    # override with vector data
    vector2 = vector.copy()
    vector2["rhi_adj"] = np.full(len(vector2), 0.8)
    vector2 = scaler.eval(vector2)
    rhi2 = vector2["rhi"]

    # override with vector attrs
    vector3 = vector.copy()
    vector3.attrs["rhi_adj"] = 0.7
    vector3 = scaler.eval(vector3)
    rhi3 = vector3["rhi"]

    # data takes priority over attrs
    vector4 = vector.copy()
    vector4.attrs["rhi_adj"] = 0.666
    vector4["rhi_adj"] = np.full(len(vector4), 0.6)
    vector4 = scaler.eval(vector4)
    rhi4 = vector4["rhi"]

    np.testing.assert_allclose(rhi0, 0.9 * rhi1, atol=1e-15)
    np.testing.assert_allclose(rhi0, 0.8 * rhi2, atol=1e-15)
    np.testing.assert_allclose(rhi0, 0.7 * rhi3, atol=1e-15)
    np.testing.assert_allclose(rhi0, 0.6 * rhi4, atol=1e-15)


def test_rhi_already_exists_warning(vector: GeoVectorDataset):
    """Confirm a warning is issued if humidity appears to have been scaled twice."""
    scaler = hs.HumidityScalingByLevel(copy_source=False)
    scaler.eval(vector)
    assert "rhi" in vector
    assert "air_pressure" in vector

    scaler = hs.ExponentialBoostLatitudeCorrectionHumidityScaling()
    with pytest.warns(UserWarning, match="Variable 'rhi' already found on source to be scaled."):
        scaler.eval(vector)


@pytest.mark.parametrize("scaler_cls", cls_list)
def test_description(scaler_cls: type):
    """Check description for scalers and ensure JSON serialization via NumpyEncoder."""
    scaler = scaler_cls()
    description = scaler.description
    assert "name" in description
    assert "formula" in description

    # other fields are numeric corresponding to parameters
    for k, v in description.items():
        if k in ["name", "formula"]:
            assert isinstance(v, str)
        else:
            assert isinstance(v, float)

    with pytest.raises(TypeError, match="serializable"):
        json.dumps(scaler)

    out = json.dumps(scaler, cls=NumpyEncoder)
    assert out == json.dumps(description)


def test_pin_specific_description():
    """Pin description for HumidityscalementConstantExponentialBoost (default for models)."""
    scaler = hs.ExponentialBoostHumidityScaling()
    assert scaler.description == {
        "name": "exponential_boost",
        "formula": "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent",
        "rhi_adj": 0.97,
        "rhi_boost_exponent": 1.7,
        "clip_upper": 1.7,
    }


@pytest.mark.parametrize("exp", [1.1, 1.2, 1.9, 2.0, 3.1])
def test_rhi_boost_exponential(vector: GeoVectorDataset, exp: float):
    """Test exponential boosting patterns."""
    base = DefaultHumidityScaling()
    _, rhi1 = base.scale(
        vector["specific_humidity"], vector["air_temperature"], vector.air_pressure
    )

    assert "rhi" not in vector
    scaler = hs.ExponentialBoostHumidityScaling(
        rhi_adj=1, rhi_boost_exponent=exp, copy_source=False
    )
    scaler.eval(vector)
    rhi2 = vector["rhi"]
    atol = 1e-15
    assert np.all(rhi2 + atol >= rhi1)

    # confirm some values above 1.0 and some below 1.0
    assert np.any(rhi2 > 1)
    assert np.any(rhi2 < 1)
    assert np.all(rhi2 <= scaler.params["clip_upper"])

    # no boosting if rhi < 1
    filt = rhi2 < 1
    np.testing.assert_allclose(rhi1[filt], rhi2[filt], atol=atol)

    # and boosting if rhi >= 1
    expected = np.minimum(rhi1[~filt] ** exp, scaler.params["clip_upper"])
    np.testing.assert_allclose(expected, rhi2[~filt], atol=atol)


@pytest.mark.parametrize("scaler_cls", cls_list)
def test_scalers_pass_nan_through(vector: GeoVectorDataset, scaler_cls: type):
    """Confirm each scaler pass nan values through when computing rhi."""
    vector["specific_humidity"][55] = np.nan
    vector["air_temperature"][66] = np.nan
    scaler = scaler_cls(copy_source=False)
    scaler.eval(vector)
    assert np.flatnonzero(np.isnan(vector["rhi"])).tolist() == [55, 66]

    # And, because some scalers convert from rhi back to q, we now get a nan
    # value there arising from the nan value in air temperature
    boosters = (
        hs.ExponentialBoostHumidityScaling,
        hs.ExponentialBoostLatitudeCorrectionHumidityScaling,
    )
    if isinstance(scaler, boosters):
        assert np.isnan(vector["specific_humidity"][66])
    else:
        assert np.isfinite(vector["specific_humidity"][66])


@pytest.fixture
def custom():
    """Create custom VectorDataset for testing ExponentialBoostLatitudeCorrectionHumidityScaling."""
    # Data from Roger: uncorrected and correct RHi values
    latitude = [2.4970, 14.4288, 48.8538, 61.7205, 69.2504, -33.4201, 45.0802, 0.6897]
    air_temperature = [200.0, 210.0, 220.0, 225.0, 230.0, 235.0, 240.0, 245.0]
    rhi_uncor = [1.070010, 1.101309, 1.084645, 1.061042, 1.379778, 1.145034, 0.394136, 0.576414]
    rhi_cor = [1.165983, 1.248463, 1.203655, 1.167633, 1.474444, 1.195842, 0.412155, 0.567973]

    vector = VectorDataset()
    vector["latitude"] = latitude
    vector["rhi_uncor"] = rhi_uncor
    vector["rhi_cor"] = rhi_cor
    vector["T"] = air_temperature

    # Get into form expected by humidity scaler: calculate q
    vector["p"] = np.full(len(vector), 25000.0)
    rhi_on_q = vector["p"] * (constants.R_v / constants.R_d) / thermo.e_sat_ice(vector["T"])
    vector["q"] = vector["rhi_uncor"] / rhi_on_q

    return vector


def test_global_rhi_correction(custom: VectorDataset):
    """Check `ExponentialBoostLatitudeCorrectionHumidityScaling` implementation."""
    scaler = hs.ExponentialBoostLatitudeCorrectionHumidityScaling()

    with pytest.raises(KeyError, match="latitude"):
        scaler.scale(custom["q"], custom["T"], custom["p"], **scaler.params)

    _, rhi_cor = scaler.scale(
        custom["q"],
        custom["T"],
        custom["p"],
        latitude=custom["latitude"],
        **scaler.params,
    )
    np.testing.assert_allclose(rhi_cor, custom["rhi_cor"], atol=1e-5)



"""Test flight module."""

from __future__ import annotations

import json
import pathlib
from typing import Any

import matplotlib.axes
import numpy as np
import pandas as pd
import pytest
from pyproj import Geod

from pycontrails import Aircraft, Flight, GeoVectorDataset, MetDataArray, MetDataset, VectorDataset
from pycontrails.core import flight
from pycontrails.models.issr import ISSR
from pycontrails.physics import constants, units

from .conftest import get_static_path

##########
# Fixtures
##########


@pytest.fixture(scope="module")
def fl(flight_data: pd.DataFrame, flight_attrs: dict[str, Any]) -> Flight:
    flight_data["constant_numeric_column"] = 1729
    flight_data["constant_poorly_formatted_str_column"] = "   contrails  "
    return Flight(flight_data, attrs=flight_attrs)


@pytest.fixture(scope="module")
def rng():
    return np.random.default_rng(seed=31415)


@pytest.fixture(scope="module")
def random_flight_list(rng: np.random.Generator) -> list[Flight]:
    n = 1000

    def _random_flight() -> Flight:
        return Flight(
            latitude=90 * rng.random(n),
            longitude=180 * rng.random(n),
            altitude=11000 * rng.random(n),
            time=pd.date_range("2019-05-31 00:00:00", "2019-05-31 00:59:00", periods=n),
        )

    return [_random_flight() for _ in range(100)]


##########
# Tests
##########


def test_flight_init(fl: Flight) -> None:
    assert isinstance(fl, GeoVectorDataset)
    assert isinstance(fl.attrs, dict)
    assert isinstance(fl.data, dict)

    assert "longitude" in fl.data
    assert "latitude" in fl.data
    assert "altitude" in fl.data
    assert "time" in fl.data
    assert "constant_numeric_column" in fl.data

    assert "equip" in fl.attrs

    assert isinstance(fl.level, np.ndarray)
    assert isinstance(fl.air_pressure, np.ndarray)


def test_flight_crs(flight_data: pd.DataFrame) -> None:
    fl = Flight(flight_data)

    # crs defaults to "EPSG:4326"
    assert fl.attrs["crs"] == "EPSG:4326"

    # crs maintained as attr
    fl = Flight(flight_data, crs="EPSG:3857")
    assert fl.attrs["crs"] == "EPSG:3857"


def test_flight_creation() -> None:
    """Test Flight.__init__() with unsorted time data."""
    # sorted
    longitude = np.array([0, 1, 2])
    latitude = np.array([0, 1, 2])
    altitude = np.array([11000, 11200, 11300])
    time = np.array(
        [
            np.datetime64("2021-09-01 00:00:00"),
            np.datetime64("2021-08-31 23:00:00"),
            np.datetime64("2021-09-01 01:00:00"),
        ]
    )

    # warn if copy=True and time not sorted
    with pytest.warns(UserWarning, match="Sorting Flight data by time."):
        fl = Flight(longitude=longitude, latitude=latitude, altitude=altitude, time=time)

    assert fl["latitude"][0] == latitude[1]
    assert fl["latitude"][1] == latitude[0]
    assert fl["latitude"][2] == latitude[2]

    assert fl["longitude"][0] == longitude[1]
    assert fl["longitude"][1] == longitude[0]
    assert fl["longitude"][2] == longitude[2]

    assert fl["altitude"][0] == altitude[1]
    assert fl["altitude"][1] == altitude[0]
    assert fl["altitude"][2] == altitude[2]

    assert fl["time"][0] == time[1]
    assert fl["time"][1] == time[0]
    assert fl["time"][2] == time[2]

    # raise if copy=False and time not sorted
    with pytest.raises(ValueError, match="must be sorted"):
        fl = Flight(
            longitude=longitude, latitude=latitude, altitude=altitude, time=time, copy=False
        )


def test_flight_aircraft(flight_data: pd.DataFrame) -> None:
    aircraft = Aircraft(aircraft_type="A380")

    # aircraft properties set to attrs, including None properties
    fl = Flight(flight_data, aircraft=aircraft)
    assert fl.attrs["aircraft_type"] == "A380"
    assert fl.attrs["wingspan"] is None

    # input attrs should override
    fl = Flight(flight_data, aircraft=aircraft, attrs={"aircraft_type": "B777"})
    assert fl.attrs["aircraft_type"] == "B777"
    assert fl.attrs["wingspan"] is None

    fl = Flight(flight_data, aircraft=aircraft, aircraft_type="B777")
    assert fl.attrs["aircraft_type"] == "B777"
    assert fl.attrs["wingspan"] is None


def test_flight_constants(fl: Flight) -> None:
    assert isinstance(fl.constants, dict)
    assert fl.constants["equip"] == "A359"
    assert fl.constants["flight"] == "QR172"
    assert fl.constants["real_to"] == fl.constants["schd_to"]
    assert fl.constants["constant_numeric_column"] == 1729
    assert fl.constants["constant_poorly_formatted_str_column"] == "contrails"


def test_flight_properties(fl: Flight) -> None:
    # test level
    assert isinstance(fl.level, np.ndarray)
    assert np.all(fl.level == units.m_to_pl(fl["altitude"]))

    # test air_pressure
    assert isinstance(fl.air_pressure, np.ndarray)
    assert np.all(fl.air_pressure == 100 * fl.level)

    # constants tested separates in `test_flight_constants`

    # test coords
    assert isinstance(fl.coords, dict)
    assert "longitude" in fl.coords and fl.coords["longitude"] is fl.data["longitude"]
    assert "latitude" in fl.coords and fl.coords["latitude"] is fl.data["latitude"]
    assert "time" in fl.coords and fl.coords["time"] is fl.data["time"]
    assert "level" in fl.coords and np.all(fl.coords["level"] == fl.level)


def test_flight_hashable(fl: Flight) -> None:
    assert isinstance(fl.hash, str)
    assert fl.hash == "f78fcb6235a6ba3a0e0dc418034736b2a3618720"


def test_flight_empty() -> None:
    fl = Flight.create_empty(attrs=dict(equip="A359"))

    assert isinstance(fl.data, dict)
    assert "longitude" in fl.data and len(fl.data["longitude"]) == 0
    assert "latitude" in fl.data and len(fl.data["latitude"]) == 0
    assert "altitude" in fl.data and len(fl.data["altitude"]) == 0
    assert "time" in fl.data and len(fl.data["time"]) == 0
    assert len(fl.air_pressure) == 0
    assert len(fl.level) == 0

    assert fl.attrs["crs"] == "EPSG:4326"
    assert isinstance(fl.constants, dict) and fl.constants == fl.attrs

    assert not fl


def test_flight_validate(flight_data: pd.DataFrame) -> None:
    """Test altitude warning and coordinate errors in Flight __init__."""
    df = flight_data.copy()
    df["altitude"] = 17000
    match = (
        "Flight altitude is high. Expected altitude unit is meters. "
        "Found waypoint with altitude 17000 m."
    )
    with pytest.warns(UserWarning, match=match):
        Flight(df)

    match = (
        "Flight altitude is high for flight xyz. Expected altitude unit is meters. "
        "Found waypoint with altitude 17000 m."
    )
    with pytest.warns(UserWarning, match=match):
        Flight(df, flight_id="xyz")

    df = flight_data.copy()
    df["longitude"] = 200
    with pytest.raises(ValueError, match="EPSG:4326 longitude coordinates"):
        Flight(df)

    df = flight_data.copy()
    df["latitude"] = 101
    with pytest.raises(ValueError, match="EPSG:4326 latitude coordinates"):
        Flight(df)


def test_flight_fitting() -> None:
    df = pd.read_parquet(get_static_path("flight-spire-data-cleaning.pq"))
    df.rename(columns={"altitude_baro": "altitude_ft", "timestamp": "time"}, inplace=True)
    df["time"] = df["time"].dt.tz_localize(None)
    flight = Flight(df[df["callsign"] == "BAW506"], drop_duplicated_times=True)
    f_copy = flight.copy()
    smoothed_flight = flight.fit_altitude()

    assert abs(smoothed_flight["altitude_ft"][1100] - 37000) < 1e-6
    assert abs(smoothed_flight["altitude_ft"][1200] - 37000) < 1e-6
    assert abs(smoothed_flight["altitude_ft"][1300] - 37000) < 1e-6
    assert f_copy == flight


def test_flight_eq(fl: Flight) -> None:
    # classes should be equal
    fl2 = Flight(fl.data, attrs=fl.attrs)
    fl3 = Flight(fl.data, attrs=fl.attrs)
    assert fl2.data is not fl3.data
    assert fl2.attrs is not fl3.attrs
    assert fl2 == fl3

    # data should be compared elementwise
    fl4 = Flight(fl.data, attrs=fl.attrs)
    fl4.data["altitude"][3] = 10
    assert fl2 != fl4

    # attributes with ndarrays should be compared elementwise
    fl2.attrs["np attr"] = np.array([1, 2, 3])
    fl3.attrs["np attr"] = np.array([1, 2, 3])
    assert fl2 == fl3

    fl5 = Flight(fl.data, attrs=fl.attrs)
    fl5.attrs["np attr"] = np.array([1, 2, 4])
    assert fl2 != fl5

    # nan values okay
    fl2.data["altitude"][3] = np.nan
    fl3.data["altitude"][3] = np.nan
    assert fl2 == fl3


def test_flight_repr(fl: Flight) -> None:
    s = repr(fl)
    assert isinstance(s, str)
    assert "Keys:" in s
    assert "Attributes:" in s
    assert "time" in s
    assert "longitude" in s
    assert "latitude" in s
    assert "altitude" in s
    assert "equip" in s
    assert "flight" in s
    assert "real_to" in s
    assert "constant_numeric_column" not in s
    for key in fl.data:
        assert key in fl._repr_html_()


def test_ensure_vars(fl: Flight) -> None:
    assert fl.ensure_vars(["time", "latitude", "longitude", "equip", "flight"])
    with pytest.raises(KeyError, match="Flight instance does not contain data or attr"):
        fl.ensure_vars(["time2"])

    is_var = fl.ensure_vars(["time2"], raise_error=False)
    assert not is_var


def test_plot(fl: Flight) -> None:
    ax = fl.plot()
    assert isinstance(ax, matplotlib.axes.Axes)
    assert ax.get_xlabel() == "longitude"
    assert ax.get_ylabel() == "latitude"

    # kwargs
    assert isinstance(fl.plot(figsize=(20, 20)), matplotlib.axes.Axes)
    assert isinstance(fl.plot(alpha=0.5), matplotlib.axes.Axes)
    assert isinstance(fl.plot(marker="o", linestyle=""), matplotlib.axes.Axes)


def test_flight_time_methods(fl: Flight) -> None:
    assert isinstance(fl.duration, pd.Timedelta)
    assert isinstance(fl.max_time_gap, pd.Timedelta)
    assert fl.max_time_gap < fl.duration


def test_flight_length_methods(fl: Flight) -> None:
    assert isinstance(fl.length, float)
    x0, y0 = fl.data["longitude"][0], fl.data["latitude"][0]
    x1, y1 = fl.data["longitude"][-1], fl.data["latitude"][-1]

    # comparing length against geodesic from start to end
    geod = Geod(ellps="WGS84")
    crow_flies_dist = geod.line_length([x0, x1], [y0, y1]) / 1000
    assert crow_flies_dist <= fl.length

    assert isinstance(fl.max_distance_gap, float)
    assert fl.max_distance_gap < fl.length


def test_flight_filtering_methods(flight_data: pd.DataFrame, flight_attrs: dict[str, Any]) -> None:
    data = flight_data.copy()
    data.loc[100] = data.loc[101]
    data.loc[51] = data.loc[52]
    with pytest.warns(UserWarning, match="Flight contains 2 duplicate times"):
        fl = Flight(data=data, attrs=flight_attrs)
    fl2 = fl.filter_by_first()
    assert fl.attrs == fl2.attrs
    assert fl.duration == fl2.duration
    assert fl.max_time_gap <= fl2.max_time_gap
    assert len(fl) - len(fl2) == 2
    for col in fl.data:
        assert set(fl.data[col]) == set(fl2.data[col])

    fl3 = fl2.resample_and_fill("10S")
    assert fl3.max_time_gap == pd.Timedelta("10S")
    assert np.all(np.diff(fl3.data["time"]) == pd.Timedelta("10S"))
    assert set(fl3.data.keys()) == {"time", "latitude", "longitude", "altitude"}


def test_resampling(fl: Flight, flight_meridian: Flight) -> None:
    fl2 = fl.resample_and_fill("10S")
    fl3 = fl.resample_and_fill("10T")
    # flight with more waypoints will waver more, and have longer length
    assert len(fl2) > len(fl3)

    # fl2 has 60 times more data
    assert abs(len(fl3) - len(fl2) / 60) <= 1

    # at large threshold, geodesics are not calculated
    fl4 = fl.resample_and_fill("1T", "linear")
    fl5 = fl.resample_and_fill("1T", "geodesic", 1000e3)
    assert fl4 == fl5

    # test resample over the meridian
    # all resampled waypoints should be < 1 deg of longitude apart on [0, 360] scale
    fl6 = flight_meridian.resample_and_fill("1T")
    assert np.all(np.abs(np.diff(fl6["longitude"] % 360)) < 1)

    fl7 = flight_meridian.resample_and_fill("3T")
    assert np.all(np.abs(np.diff(fl7["longitude"] % 360)) < 1)

    fl8 = flight_meridian.resample_and_fill("5T")
    assert np.all(np.abs(np.diff(fl8["longitude"] % 360)) < 2)


def test_geodesic_interpolation(fl: Flight) -> None:
    fl2 = fl.resample_and_fill("1T", "geodesic", geodesic_threshold=1e3)
    fl3 = fl.resample_and_fill("1T", "geodesic", geodesic_threshold=200e3)
    # fine geodesic interpolation will beat linear interpolation for minimizing spherical distance
    assert fl2.length < fl3.length
    with pytest.raises(ValueError, match="Unknown `fill_method`"):
        fl.resample_and_fill(fill_method="nearest")


def test_altitude_interpolation(fl: Flight) -> None:
    # check default flight
    fl1 = fl.resample_and_fill("10S")
    fl2 = fl.resample_and_fill("10T")

    def _check_rocd(_fl: Flight, nominal_rocd: float = constants.nominal_rocd) -> np.bool_:
        """Check rate of climb/descent."""
        dt = np.diff(_fl["time"], append=np.datetime64("NaT")) / np.timedelta64(1, "s")
        dalt = np.diff(_fl.altitude, append=np.nan)
        rocd = np.abs(dalt / dt)
        return np.all(rocd[:-1] < 2 * nominal_rocd)

    # confirm that all d_altitude values are < 2 * the nominal rate of climb/descent
    assert _check_rocd(fl1)
    assert _check_rocd(fl2)

    # test more aggresive altitude interpolation
    fl_alt = Flight(
        longitude=np.linspace(0, 10, 10),
        latitude=np.linspace(0, 10, 10),
        time=pd.date_range("2000-01-01 00:00:00", "2000-01-01 05:00:00", periods=10),
        altitude=np.array([0, 11000, 5000, 9000, 9000, 11000, 5000, 11000, 11000, 0]),
    )

    # resample to 1 min
    fl10 = fl_alt.resample_and_fill("1T")

    # confirm that rocd is appropriate
    assert _check_rocd(fl10)

    # resample to 5 minutes
    # confirm that all values exist in previous resampling
    fl11 = fl_alt.resample_and_fill("5T")
    assert _check_rocd(fl11)
    assert np.in1d(fl11["altitude"], fl10["altitude"]).all()

    # test altitude interpolation with level
    fl_lev = Flight(
        longitude=np.linspace(0, 10, 10),
        latitude=np.linspace(0, 10, 10),
        time=pd.date_range("2000-01-01 00:00:00", "2000-01-01 05:00:00", periods=10),
        level=np.array([1000, 200, 300, 300, 250, 250, 300, 200, 200, 1000]),
    )

    # resample to 1 min
    fl13 = fl_lev.resample_and_fill("1T")
    assert "level" not in fl13
    assert _check_rocd(fl13)

    # test nominal rocd
    fl14 = fl_alt.resample_and_fill("1T", nominal_rocd=30)
    assert _check_rocd(fl14, nominal_rocd=30)

    # test warning with low nominal rocd
    with pytest.warns(UserWarning, match="Rate of climb/descent values greater than nominal"):
        fl15 = fl_alt.resample_and_fill("1T", nominal_rocd=1)
        assert not _check_rocd(fl15, nominal_rocd=1)

    # test `drop` kwarg
    fl_alt["extrakey"] = np.linspace(0, 10, 10)
    fl_alt["level"] = fl_alt.level
    fl16 = fl_alt.resample_and_fill("1T", drop=False)
    assert "extrakey" in fl16 and np.any(np.isnan(fl16["extrakey"]))
    assert "level" not in fl16

    fl17 = fl_alt.resample_and_fill("1T")
    assert "extrakey" not in fl17


def test_geojson_methods(fl: Flight, rng: np.random.Generator) -> None:
    d1 = fl.to_geojson_points()
    assert d1["type"] == "FeatureCollection"
    for feature in d1["features"]:
        assert feature["geometry"]["type"] == "Point"
        assert len(feature["geometry"]["coordinates"]) == 3
    assert json.dumps(d1)

    d2 = fl.to_geojson_linestring()
    assert d2["type"] == "FeatureCollection"
    assert json.dumps(d2)
    assert len(d1["features"]) == len(d2["features"][0]["geometry"]["coordinates"])
    for feature, coord in zip(d1["features"], d2["features"][0]["geometry"]["coordinates"]):
        assert feature["geometry"]["coordinates"] == coord

    with pytest.raises(KeyError):
        fl.to_geojson_multilinestring("bad key")
    fl["issr"] = rng.integers(0, 2, len(fl))
    d3 = fl.to_geojson_multilinestring("issr")
    assert d3["type"] == "FeatureCollection"
    assert len(d3["features"]) == 2

    coords1 = [f["geometry"]["coordinates"] for f in d3["features"]]
    # flattening, casting to tuple to allow inclusion into a set
    coords1 = [tuple(c) for c2 in coords1 for c1 in c2 for c in c1]
    coords2 = [tuple(c) for c in d2["features"][0]["geometry"]["coordinates"]]
    assert set(coords1) == set(coords2)


# ignoring the warnings that come with traffic
# NOTE: this does not work on docker, so we skip
def _is_docker() -> bool:
    path1 = pathlib.Path("/proc/self/cgroup")
    path2 = pathlib.Path("/.dockerenv")
    return (path1.is_file() and any("docker" in line for line in open(path1))) or path2.is_file()


def _is_traffic_installed() -> bool:
    try:
        import traffic  # noqa: F401
    except ModuleNotFoundError:
        return False
    return True


@pytest.mark.skipif(not _is_traffic_installed(), reason="Traffic is no longer optional dependency")
def test_to_traffic(fl: Flight) -> None:
    tr = fl.to_traffic()
    # NOTE: tf and fl have inconsistent column names (time and timestamp), so just ensuring
    # the values were copied correctly on tr instantiation.
    np.testing.assert_array_equal(fl.dataframe.to_numpy(), tr.data.to_numpy())
    assert fl.duration == tr.duration


def test_crs(fl: Flight, met_issr: MetDataset, rng: np.random.Generator) -> None:
    fl2 = fl.to_pseudo_mercator()
    assert fl.attrs["crs"] == "EPSG:4326"
    assert fl2.attrs["crs"] == "EPSG:3857"

    assert np.all(fl.data["altitude"] == fl2.data["altitude"])
    assert np.all(fl.data["time"] == fl2.data["time"])
    assert np.all(fl.data["latitude"] != fl2.data["latitude"])
    assert np.all(fl.data["longitude"] != fl2.data["longitude"])

    with pytest.raises(NotImplementedError):
        fl2.length
    fl2.update(issr=rng.integers(0, 2, len(fl2)))
    with pytest.raises(NotImplementedError, match="Only implemented for EPSG:4326"):
        fl2.length_met("issr")
    with pytest.raises(NotImplementedError, match="Only implemented for EPSG:4326"):
        fl2.proportion_met("issr")
    with pytest.raises(AttributeError, match="has no attribute 'interpolate'"):
        fl2.intersect_met(met_issr)

    with pytest.raises(KeyError, match="Column key does not exist in data"):
        fl.length_met("key")  # key not in fl.data


def test_antimeridian_jump() -> None:
    df = pd.DataFrame(
        {
            "longitude": [-177, -179, 179, 178],
            "latitude": [40, 41, 42, 41],
            "time": pd.date_range("2000-01-01 00:00:00", "2000-01-01 00:30:00", periods=4),
            "altitude": [10000, 10001, 10003, 10006],
            "issr": [0, 1, 1, 1],
        }
    )

    fl = Flight(df)
    d = fl.to_geojson_multilinestring("issr", split_antimeridian=True)
    issr_feature = d["features"][1]
    assert len(issr_feature["geometry"]["coordinates"]) == 2

    d = fl.to_geojson_multilinestring("issr", split_antimeridian=False)
    issr_feature = d["features"][1]
    assert len(issr_feature["geometry"]["coordinates"]) == 1

    fl = fl.to_pseudo_mercator()
    d = fl.to_geojson_multilinestring("issr", split_antimeridian=True)
    issr_feature = d["features"][1]
    assert len(issr_feature["geometry"]["coordinates"]) == 2

    d = fl.to_geojson_multilinestring("issr", split_antimeridian=False)
    issr_feature = d["features"][1]
    assert len(issr_feature["geometry"]["coordinates"]) == 1

    df["longitude"] = [-177, -179, 179, -178]
    fl = Flight(df)
    # jumps antimeridian twice
    with pytest.raises(ValueError):
        fl.to_geojson_multilinestring("issr", split_antimeridian=True)


def test_segment_properties(fl: Flight, rng: np.random.Generator) -> None:
    # segment duration
    segment_duration = fl.segment_duration()
    assert len(segment_duration) == len(fl)
    assert np.isnan(segment_duration[-1])
    assert 1 <= np.nanmax(segment_duration) <= 198

    # segment lengths
    segment_length = fl.segment_length()
    assert len(segment_length) == len(fl)
    assert np.isnan(segment_length[-1])
    assert np.nanmax(segment_length) <= 5.1e4

    # segment angles
    sin_a, cos_a = fl.segment_angle()
    assert len(sin_a) == len(fl)
    assert len(cos_a) == len(fl)
    assert np.all(sin_a[:-1] <= 1)
    assert np.all(sin_a[:-1] >= -1)
    assert np.all(cos_a[:-1] <= 1)
    assert np.all(cos_a[:-1] >= -1)
    assert np.isnan(sin_a[-1])
    assert np.isnan(cos_a[-1])

    # segment ground airspeed
    segment_groundspeed = fl.segment_groundspeed()
    assert len(segment_groundspeed) == len(fl)
    assert np.nanmin(segment_groundspeed) > 0 and np.nanmax(segment_groundspeed) <= 300
    assert np.isnan(segment_groundspeed[-1])

    # segment true airspeed
    segment_true_airspeed = fl.segment_true_airspeed(smooth=False)
    assert len(segment_true_airspeed) == len(fl)
    assert np.isnan(segment_true_airspeed[-1])
    abs_diff = np.abs(segment_true_airspeed - segment_groundspeed)
    # should be similar with no wind speed
    assert np.all(abs_diff[:-1] < 1)

    segment_true_airspeed = fl.segment_true_airspeed()
    assert len(segment_true_airspeed) == len(fl)
    assert np.isnan(segment_true_airspeed[-1])
    abs_diff = np.abs(segment_true_airspeed - segment_groundspeed)
    assert np.all(abs_diff[:-1] <= 50)

    u_wind = rng.random(len(fl)) * 10
    v_wind = rng.random(len(fl)) * 5
    segment_true_airspeed = fl.segment_true_airspeed(u_wind=u_wind, v_wind=v_wind)
    assert len(segment_true_airspeed) == len(fl)
    assert np.isnan(segment_true_airspeed[-1])
    assert np.all(abs_diff[:-1] <= 50)

    # mach number
    air_temperature = rng.random(len(fl)) * 30 + 273
    segment_mach_number = fl.segment_mach_number(segment_true_airspeed, air_temperature)
    assert len(segment_mach_number) == len(fl)
    assert np.all(segment_mach_number[:-1] < 2)


def test_downselect_met(fl: Flight, met_issr: MetDataset) -> None:
    # met_issr is session scoped and we're about to mutate it
    met = met_issr.copy()

    # adjust met level to get more overlap
    np.testing.assert_array_equal(met["level"].values, [225, 250, 300])
    met.update(level=[150, 300, 500])
    np.testing.assert_array_equal(met["level"].values, [150, 300, 500])
    np.testing.assert_array_equal(met["air_temperature"].data["level"], [150, 300, 500])

    # clip flight domain so it is definitely smaller than met
    mask = (fl.level < 500) & (fl.level > 150)
    fl = fl.filter(mask)

    assert tuple(met.data.sizes.values()) == (15, 8, 3, 2)
    met_downselected = fl.downselect_met(met)

    # original not mutated
    assert tuple(met.data.sizes.values()) == (15, 8, 3, 2)
    assert tuple(met_downselected.data.sizes.values()) == (3, 3, 3, 2)
    assert met_downselected is not met
    # name convenience
    ds = met_downselected.data

    # check the downselection
    assert ds["longitude"].max() == 75
    assert ds["longitude"].values[-2] < fl["longitude"].max() < ds["longitude"].values[-1]
    assert ds["latitude"].max() == 65
    assert ds["latitude"].values[-2] < fl["latitude"].max() < ds["latitude"].values[-1]
    assert ds["level"].max() == 500
    assert ds["level"].values[-2] < fl.level.max() < ds["level"].values[-1]
    assert ds["time"].max() == np.datetime64("2019-05-31T06:00:00")

    assert ds["longitude"].min() == 25
    assert ds["longitude"].values[0] < fl["longitude"].min() < ds["longitude"].values[1]
    assert ds["latitude"].min() == 15
    assert ds["latitude"].values[0] < fl["latitude"].min() < ds["latitude"].values[1]
    assert ds["level"].min() == 150
    assert ds["level"].values[0] < fl.level.min() < ds["level"].values[1]
    assert ds["time"].min() == np.datetime64("2019-05-31T05:00:00")
    assert ds["time"].values[0] < fl["time"].min() < ds["time"].values[1]


@pytest.mark.parametrize("freq", ["10S", "1T", "10T"])
@pytest.mark.parametrize("shift", [-180, 0])
def test_interpolation_edge_cases(shift: int, freq: str) -> None:
    """Test interpolation adjustments for antimeridian crossing.

    This test probes both the core logic of `resample_and_fill` as well as
    the logic in `_altitude_interpolation`.

    The parameter `shift` translates the longitude coordinates. In the case of
    `shift = 0` and `shift=6`, the flight crosses the antimeridian.

    The parameter `freq` various widely, allowing for both up and down sampling.
    """
    # Using level with many missing values
    fl = Flight(
        longitude=(np.arange(-10, 10, dtype=float) % 360 + shift) % 360 - 180,
        latitude=np.arange(20),
        level=np.r_[200, [np.nan] * 5, 300, 305, 309, 313, 311, [np.nan] * 7, 189, 188],
        time=pd.date_range("2022-01-01", "2022-01-01T03", 20),
    )
    fl["longitude"][3] = np.nan  # make test more interesting
    fl2 = fl.resample_and_fill(freq)
    jump = np.nonzero(np.diff(fl2["longitude"]) < 0)[0]
    if shift == -180:
        assert jump.size == 0
    elif shift == 0:
        assert jump.size == 1

    # Interpolated altitudes should increase then decrease
    np.diff(np.sign(np.diff(fl2.altitude))).nonzero()[0].size == 1

    assert not fl2.dataframe.isna().any().any()


@pytest.mark.parametrize("direction", ["east", "west"])
@pytest.mark.parametrize("cross_anti", [True, False])
def test_antimeridian_long_cross(direction: str, cross_anti: bool) -> None:
    """Test interpolation adjustments for antimeridian crossing.

    This test uses longitude values spanning at least three quadarants AND
    crossing the antimeridian.
    """
    # Intentionally keep the logic very boilerplate
    # In all cases, d_lon = 170
    n = 100
    if direction == "east" and cross_anti:
        longitude = np.linspace(80, 80 + 170, n)
    elif direction == "east" and not cross_anti:
        longitude = np.linspace(-40, -40 + 170, n)
    elif direction == "west" and cross_anti:
        longitude = np.linspace(190, 190 - 170, n)
    elif direction == "west" and not cross_anti:
        longitude = np.linspace(60, 60 - 170, n)
    else:
        raise ValueError("Unknown param")
    longitude = (longitude + 180) % 360 - 180
    fl = Flight(
        longitude=longitude,
        latitude=np.zeros(n),
        altitude=np.full(n, 10000),
        time=pd.date_range("2022-01-01", "2022-01-01T03", n),
    )
    fl2 = fl.resample_and_fill("5S")
    assert np.all(fl2["longitude"] < 180)
    assert np.all(fl2["longitude"] >= -180)
    assert np.all(np.isfinite(fl2["longitude"]))
    assert np.all(fl2.segment_length()[:-1] < 10000)
    assert np.all(fl2.segment_length()[:-1] > 8000)


def test_intersect_issr_met(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test `intersect_met` and `length_met` methods."""
    issr = ISSR(met_era5_fake).eval()
    fl = flight_fake.copy()
    fl["issr"] = fl.intersect_met(issr)

    # at least one waypoint with ISSR
    assert np.nansum(fl["issr"])

    # somes ISSR values are interpolated between 0 and 1
    assert np.any((fl["issr"] > 0) & (fl["issr"] < 1))
    with pytest.warns(UserWarning, match="Column issr contains real numbers between 0 and 1."):
        fl.length_met("issr")

    # relaxing threshold
    assert fl.length_met("issr", threshold=0.5)

    # waypoints below lowest-altitude level in era5 get filled with NaN
    low_altitude_waypoints = fl["altitude"] < met_era5_fake.data["altitude"].min().item()
    assert np.isnan(fl["issr"][low_altitude_waypoints]).all()
    assert ~np.isnan(fl["issr"][~low_altitude_waypoints]).all()

    # based on interpolation, we end up with an issr value exactly when flight waypoint
    # longitude value is within 1 of a multiple of 5
    lons = fl["longitude"][~low_altitude_waypoints]
    expect_issr = np.abs((lons + 2.5) % 5 - 2.5) < 1
    actual_issr = fl["issr"][~low_altitude_waypoints].astype(bool)
    assert np.all(actual_issr == expect_issr)


def test_intersect_met_method_nearest(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test `intersect_met` with "nearest" interpolation."""
    fl = flight_fake.copy()

    # results agree when calling intersect_met and intersect_model with method='nearest'
    issr = ISSR(met_era5_fake).eval()
    fl["issr"] = fl.intersect_met(issr, method="nearest", fill_value=0)

    # values in issr column are either 0 or 1
    assert np.all(np.isin(fl["issr"], [0, 1]))
    assert fl["issr"][0] == 0

    # extrapolates values when "nearest" and fill_value set to None
    fl.update(issr=fl.intersect_met(issr, method="nearest", fill_value=None))
    assert np.all(np.isin(fl["issr"], [0, 1]))
    assert fl["issr"][0] == 1


def test_intersect_met_method_linear(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    """Test `intersect_met` with "linear" interpolation."""
    fl = flight_fake.copy()

    # results disagree when calling intersect_met and intersect_model with default method='linear'
    issr = ISSR(met_era5_fake).eval()
    fl["issr"] = fl.intersect_met(issr, method="linear")

    # values in fl2 issr column are floating point values
    assert (fl["issr"][~np.isnan(fl["issr"])] < 1).all()


def test_intersect_other_mets(met_era5_fake: MetDataset, flight_fake: Flight) -> None:
    fl = flight_fake.copy()

    da = met_era5_fake["specific_humidity"]
    fl["specific_humidity"] = fl.intersect_met(da)
    assert "specific_humidity" in fl.data

    # creating a meaningless DataArray that has the right dimension variables
    da = (
        da.data["longitude"]
        + da.data["altitude"]
        + da.data["latitude"]
        + 0 * da.data["time"].astype(int)
    )

    fl.data["meaningless"] = fl.intersect_met(MetDataArray(da))

    assert "specific_humidity" in fl
    assert "meaningless" in fl
    # 5 geospatial columns, specific_humidity, and meaningless
    assert len(fl.data) == 6

    cruising_altitude_waypoints = fl["altitude"] >= met_era5_fake.data["altitude"].min().item()

    # altitude is the dominate term in the meaningless column
    assert (
        (fl["meaningless"][cruising_altitude_waypoints] > 7500)
        & (fl["meaningless"][cruising_altitude_waypoints] < 13500)
    ).all()


def test_cast_to_vector(flight_fake: Flight) -> None:
    """Ensure casting works as expected."""
    # Casting down
    v1 = VectorDataset(flight_fake)
    assert isinstance(v1, VectorDataset)
    assert not isinstance(v1, Flight)
    assert v1 == flight_fake

    v1["longitude"][5] = np.nan
    assert v1 != flight_fake

    v2 = GeoVectorDataset(flight_fake)
    assert isinstance(v2, GeoVectorDataset)
    assert not isinstance(v2, Flight)
    assert v2 == flight_fake

    # Casting up
    lon = v1.data.pop("longitude")
    with pytest.raises(KeyError, match="Missing required data keys"):
        GeoVectorDataset(v1)

    v1["longitude"] = lon
    v3 = GeoVectorDataset(v1)
    assert isinstance(v3, GeoVectorDataset)
    assert v3 == v1


def test_flight_duplicated_times(flight_fake: Flight) -> None:
    """Ensure that duplicate times are removed."""
    fl = flight_fake.copy()
    fl["time"][3] = fl["time"][2]
    with pytest.warns(UserWarning, match="Flight contains 1 duplicate times"):
        Flight(fl.data)

    fl = Flight(fl.data, drop_duplicated_times=True)
    assert fl.size == flight_fake.size - 1

    # We kept the 2nd row and the 3rd is different
    pd.testing.assert_series_equal(fl.dataframe.iloc[2], flight_fake.dataframe.iloc[2])
    pd.testing.assert_series_equal(
        fl.dataframe.iloc[3], flight_fake.dataframe.iloc[4], check_names=False
    )


@pytest.mark.parametrize("kernel_size", [3, 5, 7, 9, 11, 13, 15, 17])
def test_filter_altitude(kernel_size: int) -> None:
    """Check that noise in cruise altitude is removed."""

    altitude_ft = np.array([40000, 39975, 40000, 40000, 39975, 40000, 40000, 40025, 40025, 40000])
    altitude_cleaned = flight.filter_altitude(altitude_ft, kernel_size=kernel_size)
    assert altitude_cleaned.size == altitude_ft.size

    np.testing.assert_array_equal(altitude_cleaned[:-1], 40000.0)

    # Final waypoint has altitude 40002 +/- 2 ft
    assert altitude_cleaned[-1] == pytest.approx(40002, abs=2.5)


# Compare with OpenAP implementation
try:
    from openap import FlightPhase

    OPENAP_AVAILABLE = True
except ImportError:
    OPENAP_AVAILABLE = False


@pytest.mark.skipif(not OPENAP_AVAILABLE, reason="OpenAP not installed")
def test_compare_segment_phase() -> None:
    """Compare flight phase algorithm with OpenAP.

    See https://github.com/junzis/openap/blob/master/test/test_phase.py
    """
    df = pd.read_csv(
        "https://raw.githubusercontent.com/junzis/openap/master/test/data/flight_phlab.csv"
    )

    ts0 = df["ts"].values
    ts0 = ts0 - ts0[0]
    alt0 = df["alt"].values
    spd0 = df["spd"].values
    roc0 = df["roc"].values

    ts = np.arange(0, ts0[-1], 1)
    alt = np.interp(ts, ts0, alt0)
    spd = np.interp(ts, ts0, spd0)
    roc = np.interp(ts, ts0, roc0)

    fp = FlightPhase()
    fp.set_trajectory(ts, alt, spd, roc)
    labels = fp.phaselabel()

    assert labels

    Flight(data=df.rename({"alt": "altitude", "lat": "latitude", "lon": "longitude"}))
